import { Subject, BehaviorSubject } from 'rxjs';

import { AxiosInstance } from 'axios';

import { StatusCode, FileUtil } from '@ucap/core';
import {
  NativeService,
  UpdateCheckConfig,
  UpdateInfo,
  NativeType,
  WindowState,
  WindowIdle,
  NotificationRequest,
  NotificationType
} from '@ucap/native';

import { NotificationService } from './notification.service';

export class BrowserNativeService extends NativeService {
  // tslint:disable-next-line: variable-name
  private _notificationService: NotificationService;
  // tslint:disable-next-line: variable-name
  private _idle_checker: WindowIdleChecker;

  constructor(private axios: AxiosInstance) {
    super();

    this._notificationService = new NotificationService();
  }

  platform_nativeType(): Promise<NativeType> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve(NativeType.Browser);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_networkInfo(): Promise<any> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve([{ ip: 'Browser', mac: 'browser' }]);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_execute(executableName: string): Promise<number> {
    return new Promise<number>((resolve, reject) => {
      try {
        resolve(-1);
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_openDefaultBrowser(
    url: string,
    options?: {
      name?: string;
      features?: string;
      replace?: boolean;
    }
  ): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        const name = !!options && !!options.name ? options.name : null;
        const features =
          !!options && !!options.features ? options.features : null;
        window.open(url, name, features);
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  platform_readFromClipboard(): Promise<{
    text?: string;
    rtf?: string;
    html?: string;
    image?: Buffer;
    imageDataUrl?: string;
  }> {
    return new Promise<{
      text?: string;
      rtf?: string;
      html?: string;
      image?: Buffer;
      imageDataUrl?: string;
    }>((resolve, reject) => {
      try {
        navigator.permissions
          .query({ name: 'clipboard-read' as PermissionName })
          .then((result) => {
            if ('granted' === result.state || 'prompt' === result.state) {
              navigator.clipboard
                .readText()
                .then((value) => {
                  resolve({ text: value });
                })
                .catch((reason) => {
                  reject(reason);
                });
            }
          })
          .catch((reason) => {
            reject(reason);
          });
      } catch (error) {
        reject(error);
      }
    });
  }

  file_save(
    buffer: Buffer,
    fileName: string,
    mimeType: string,
    path?: string
  ): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        FileUtil.save(buffer, fileName, mimeType)
          .then((fn) => {
            resolve(fn);
          })
          .catch((reason) => {
            reject(reason);
          });
      } catch (error) {
        reject(error);
      }
    });
  }
  file_read(path: string): Promise<Buffer> {
    return new Promise<Buffer>((resolve, reject) => {
      try {
        this.axios
          .get(path, { responseType: 'arraybuffer' })
          .then((res) => {
            resolve(Buffer.from(res.data));
          })
          .catch((reason) => reject(reason));
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openFolder(folderPath?: string, make?: boolean): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(false);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_openItem(filePath?: string): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(false);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_path(
    name: 'home' | 'documents' | 'downloads' | 'music' | 'pictures' | 'videos',
    ...appendPaths: string[]
  ): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForOpen(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    properties?: Array<
      | 'openFile'
      | 'openDirectory'
      | 'multiSelections'
      | 'showHiddenFiles'
      | 'createDirectory'
      | 'promptToCreate'
      | 'noResolveAliases'
      | 'treatPackageAsDirectory'
    >;
    message?: string;
  }): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }
  file_selectForSave(option: {
    title?: string;
    defaultPath?: string;
    filters?: {
      extensions: string[];
      name: string;
    }[];
    message?: string;
  }): Promise<{
    canceled: boolean;
    filePath: string;
  }> {
    return new Promise<{
      canceled: boolean;
      filePath: string;
    }>((resolve, reject) => {
      try {
        resolve(undefined);
      } catch (error) {
        reject(error);
      }
    });
  }

  window_onState$(): BehaviorSubject<WindowState> {
    return super.window_onState$();
  }
  window_onFocus$(): BehaviorSubject<boolean> {
    if (!this._window_onFocusSubject) {
      // tslint:disable-next-line: variable-name
      const __this = this;
      const onFocus = (event: Event) => {
        __this._window_onFocusSubject.next(true);
      };
      const onBlur = (event: Event) => {
        __this._window_onFocusSubject.next(false);
      };
      window.addEventListener('focus', onFocus);
      window.addEventListener('blur', onBlur);

      super.window_onFocus$().subscribe(
        (focus) => {},
        (error) => {},
        () => {
          window.removeEventListener('focus', onFocus);
          window.removeEventListener('blur', onBlur);
        }
      );
    }

    return super.window_onFocus$();
  }
  window_close(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_minimize(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  window_maximize(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  idle_startCheck(limitTime: number): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!!this._idle_onStateSubject) {
          return;
        }

        this._idle_checker = new WindowIdleChecker({
          limitTime,
          onIdle: () => {
            if (!!this._idle_onStateSubject) {
              this._idle_onStateSubject.next(WindowIdle.Idle);
            }
          },
          onActive: () => {
            if (!!this._idle_onStateSubject) {
              this._idle_onStateSubject.next(WindowIdle.Active);
            }
          },
          onHide: () => {
            if (!!this._idle_onStateSubject) {
              this._idle_onStateSubject.next(WindowIdle.Idle);
            }
          },
          onShow: () => {
            if (!!this._idle_onStateSubject) {
              this._idle_onStateSubject.next(WindowIdle.Active);
            }
          }
        });
        this._idle_checker.start();

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  idle_onState$(): BehaviorSubject<WindowIdle> {
    return super.idle_onState$();
  }
  idle_stopCheck(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!!this._idle_checker) {
          this._idle_checker.stop();
          this._idle_checker = undefined;
        }
        if (!!this._idle_onStateSubject) {
          this._idle_onStateSubject.complete();
          this._idle_onStateSubject = undefined;
        }

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  idle_changeLimitTime(limitTime: number): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        if (!this._idle_checker) {
          return;
        }
        this._idle_checker.changeLimitTime(limitTime);

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_version(): Promise<string> {
    return new Promise<any>((resolve, reject) => {
      try {
        resolve('');
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postInit(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogin(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this._notificationService.requestPermission();

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postLogout(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_postDestroy(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_changeAutoLaunch(autoLaunch: boolean): Promise<boolean> {
    return new Promise<boolean>((resolve, reject) => {
      try {
        resolve(true);
      } catch (error) {
        reject(error);
      }
    });
  }
  app_showNotify(req: NotificationRequest): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        this._notificationService.notify(req, () => {
          window.focus();

          switch (req.type) {
            case NotificationType.Event:
              if (!!this._chat_onOpenSubject) {
                this._chat_onOpenSubject.next(req.seq);
              }
              break;
            case NotificationType.Message:
              if (!!this._message_onOpenSubject) {
                this._message_onOpenSubject.next(req.seq);
              }
              break;
            default:
              break;
          }
        });

        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_closeAllNotify(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_checkForUpdates(currentVersion: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_startCheckForUpdate(config: UpdateCheckConfig): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_onUpdate$(): BehaviorSubject<UpdateInfo> {
    return super.app_onUpdate$();
  }
  app_stopCheckForUpdate(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_applyInstantUpdates(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }
  app_exit(): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      try {
        window.close();
        resolve();
      } catch (error) {
        reject(error);
      }
    });
  }

  app_onLogout$(): Subject<void> {
    return super.app_onLogout$();
  }
  app_onStatus$(): BehaviorSubject<StatusCode> {
    return super.app_onStatus$();
  }
  app_onShowSetting$(): Subject<void> {
    return super.app_onShowSetting$();
  }

  chat_onOpen$(): BehaviorSubject<string> {
    return super.chat_onOpen$();
  }

  message_onOpen$(): BehaviorSubject<string> {
    return super.message_onOpen$();
  }
}

interface WindowIdleCheckerConfig {
  limitTime?: number;
  events?: string[];
  onIdle?: () => void;
  onActive?: () => void;
  visibilityEvents?: string[];
  onHide?: () => void;
  onShow?: () => void;
}

const defaultWindowIdleCheckerConfig: WindowIdleCheckerConfig = {
  limitTime: 5 * 60 * 1000,
  events: ['mousemove', 'keydown', 'mousedown', 'touchstart'],
  onIdle: () => {},
  onActive: () => {},
  onHide: () => {},
  onShow: () => {},
  visibilityEvents: [
    'visibilitychange',
    'webkitvisibilitychange',
    'mozvisibilitychange',
    'msvisibilitychange'
  ]
};

const WEBKIT_HIDDEN = 'webkitHidden';
const MOZ_HIDDEN = 'mozHidden';
const MS_HIDDEN = 'msHidden';

class WindowIdleChecker {
  // tslint:disable-next-line: variable-name
  private _timerId: any;
  // tslint:disable-next-line: variable-name
  private _status: WindowIdle;
  // tslint:disable-next-line: variable-name
  private _visible: boolean;
  // tslint:disable-next-line: variable-name
  private _started: boolean;

  constructor(
    // tslint:disable-next-line: variable-name
    private _config: WindowIdleCheckerConfig
  ) {
    this._config = Object.assign({}, defaultWindowIdleCheckerConfig, _config);
    this._status = WindowIdle.Active;
    this._visible = true;
  }

  start() {
    if (!!this._timerId) {
      clearInterval(this._timerId);
    }

    this._timerId = setInterval(() => {
      if (WindowIdle.Active === this._status) {
        this._status = WindowIdle.Idle;
        this._config.onIdle();
      }
    }, this._config.limitTime);

    if (!this._started) {
      this._started = true;
    }

    this._addEventListerner(this._config.events, this._onActive);

    if (!!this._config.onShow || !!this._config.onHide) {
      this._addEventListerner(
        this._config.visibilityEvents,
        this._onVisibility
      );
    }
  }

  stop() {
    if (!!this._timerId) {
      clearInterval(this._timerId);
      this._timerId = undefined;
    }

    if (!!this._started) {
      this._removeEventListerner(this._config.events, this._onActive);

      if (!!this._config.onShow || !!this._config.onHide) {
        this._removeEventListerner(
          this._config.visibilityEvents,
          this._onVisibility
        );
      }
    }
  }

  changeLimitTime(limitTime: number) {
    this._config = {
      ...this._config,
      limitTime
    };
    this.start();
  }

  private _onActive(event: Event) {
    if (WindowIdle.Idle === this._status) {
      this._status = WindowIdle.Active;
      this._config.onActive();
    }
  }

  private _onVisibility(event: Event) {
    if (
      !!document.hidden ||
      !!document[WEBKIT_HIDDEN] ||
      !!document[MOZ_HIDDEN] ||
      !!document[MS_HIDDEN]
    ) {
      if (this._visible) {
        this._visible = false;
        this._config.onHide();
      }
    } else {
      if (!this._visible) {
        this._visible = true;
        this._config.onShow();
      }
    }
  }

  private _addEventListerner(
    events: string[],
    callback: (event: Event) => void
  ) {
    events.forEach((e) => {
      window.addEventListener(e, callback.bind(this));
    });
  }

  private _removeEventListerner(
    events: string[],
    callback: (event: Event) => void
  ) {
    events.forEach((e) => {
      window.removeEventListener(e, callback.bind(this));
    });
  }
}
